## @file Data.py
#  @author Steven Palmer and Henry Madej
#  @brief Data
#  @date 20/02/2018

from CurveADT import *
import SeqServices
from Exceptions import *


## @brief An abstract data type for storing data
class Data:
  
  MAX_SIZE = 10
  
  S = []
  Z = []
  
  ## @brief init initial data structure
  @staticmethod
  def init():
    Data.S = []
    Data.Z = []
    
  ## @brief add adds a pair to the data structure
  #  @exception throws Full exception is structure is full
  #  @exception throws IndepVarNotAscending if z is less than or equal the last
  #             value in the structure 
  #  @param s dependant variable value added to the structure
  #  @param z independent variable value aded to the structure
  @staticmethod  
  def add(s, z):
    if len(Data.S) == Data.MAX_SIZE:
      raise Full
      
    if Data.Z:
      if z <= Data.Z[-1]:
        raise IndepVarNotAscending
    
    Data.S.append(s)
    Data.Z.append(z)

  ## @brief getC gets value of the dependant variable at index i
  #  @param i index of dependant variable value
  #  @return value of dependant variable at index i
  @staticmethod
  def getC(i):
    if not (0 <= i < len(Data.S)):
      raise InvalidIndex
    
    return Data.S[i]

  ## @brief eval evaluates the a curve at a particular value of x
  #  @details  Uses linear interplation to find the value of the curve at x
  #  @param x the value to evaluate the curve at
  #  @param z value adjacent to the x value
  #  @return value of the curve at the specified value of x
  @staticmethod
  def eval(x, z):
    if not isInBounds(Data.Z, z):
      raise OutOfDomain 
    
    j = SeqServices.index(Data.Z, z)
    return interpLin(Data.Z[j], Data.S[j].eval(x), Data.Z[j+1], Data.S[j+1].eval(x), z)
    
  ## @brief slice slices a curve returning a subset of the original curve
  #  @param x the x to slice at
  #  @param i the order of the curve that is being sliced
  #  @return CurveT
  @staticmethod
  def slice(x, i):
    Y = [s.eval(x) for s in Data.S]
    
    return CurveT(Data.Z, Y, i)
