/**
 * \file GameBoard.h
 * \brief An abstract data type (ADT) for representing the state of a game of FreeCell Solitaire.
 * \author Christopher W. Schankula, MacID: schankuc
 * \date April 8th, 2018
*/

#ifndef GAMEBOARD_H
#define GAMEBOARD_H

#include <vector>
#include "CardStack.h"
#include "GameBoardTypes.h"

/**
 *  \brief Class for representing the gameboard of a game of Freecell Solitaire.
 *  \details Implements a class for representing the gameboard of a Freecell Solitaire game,
 *  as well as functions for making and validating moves and verifying win states and states with no
 *  possible moves. Implemented as an Abstract Data Type (ADT) so that client code could create and manage
 *  multiple games at once, for example in a server model.
 */
class GameBoardT {
	private:
		std::vector<CardStackT> T;
		std::vector<CardStackT> F;
		std::vector<CardStackT> H;
		std::vector<CardStackT> stackSeq(unsigned int l, unsigned int n);
		std::vector<CardStackT> tableauDeck(std::vector<CardT> deck);
		bool isValidCell(CellT c, unsigned int n);
		bool validTabTab(unsigned int t0, unsigned int t1);
		bool validTabFree(unsigned int t, unsigned int f);
		bool validTabHome(unsigned int t, unsigned int h);
		bool validFreeTab(unsigned int f, unsigned int t);
		bool validFreeFree(unsigned int f0, unsigned int f1);
		bool validFreeHome(unsigned int f, unsigned int h);
		bool oneLess(CardT c0, CardT c1);
		bool oneMore(CardT c0, CardT c1);
		bool tabPlaceable(CardT c0, CardT c1);
		bool homePlaceable(CardT c0, CardT c1);
		bool isWinningHomeCell(unsigned int i);


	public:
		/**
		 *  \brief Construct a new Freecell gameboard.
		 *  \details Create a new instance of a Freecell gameboard with all cells empty. This can be
		 *  populated manually by a client.
		 */
		GameBoardT();

		/**
		 *  \brief Construct a new Freecell gameboard from a deck of cards
		 *  \details Create a new instance of a Freecell gameboard by placing the given deck
		 *  on the tableau cells of the gameboard.
		 *  \throws invalid_argument The given initial deck does not have 52 cards.
		 */
		GameBoardT(std::vector<CardT> deck);

		/**
		 *  \brief Place a card on a given cell of the gameboard.
		 *  \details Place a card on a given cell of the gameboard. Used to manually construct an
		 *  initial game state.
		 *  \param cell The type of cell on which to place the code
		 *  \param i The identifier of the cell of type cell at which to place the new card on the gameboard.
		 *  \param card The new card to place on the top of the cell given.
		 *  \throws invalid_config The placement of the card would result in an invalid
		 *  configuration of the game board (beyond the possible sizes of the cell in question).
		 *  \throw invalid_cell The given cell does not exist on the gameboard.
		 */		
		void place(CellT cell, unsigned int i, CardT card);

		/**
		 *  \brief Determine whether a given move is valid given the current gameboard state.
		 *  \details Returns whether moving a potential card from a given cell to another given cell
		 *  is a valid move according to Freecell game rules.
		 *  \returns true if the card on the top (if any) of the given cell represented by c0 and n0 
		 *  can be moved to the given cell represented by c1 and n1. false if the move is invalid or
		 *  if the cell at c0 and n0 is empty. 
		 *  \param c0 The type of cell for the "from" cell to move the card from.
		 *  \param n0 The identifier of the cell of type cell from which to move a card.
		 *  \param c1 The type of cell for the "to" cell to move the card to.
		 *  \param n1 The identifier of the cell of type cell to which to move the card.
		 *  \throws invalid_cell One or both of the cells provided are not valid cells on the gameboard.
		 */		
		bool isValidMove(CellT c0, unsigned int n0, CellT c1, unsigned int n1);

		/**
		 *  \brief Make a move from one cell to another, if it's valid.
		 *  \details Moves the card from the top of the cell represented by c0 and n0 to the cell
		 *  represented by c1 and n1.
		 *  \param c0 The type of cell for the "from" cell to move the card from.
		 *  \param n0 The identifier of the cell of type cell from which to move a card.
		 *  \param c1 The type of cell for the "to" cell to move the card to.
		 *  \param n1 The identifier of the cell of type cell to which to move the card.
		 *  \throws invalid_cell One or both of the cells provided are not valid cells on the gameboard.
		 */	
		void makeMove(CellT c0, unsigned int n0, CellT c1, unsigned int n1);

		/**
		 *  \brief Return the cards in the cell given as a CardStackT object.
		 *  \details Returns the cards in the cell on the gameboard given as a CardStackT object.
		 *  \param cell The type of cell on which to place the code
		 *  \param i The identifier of the cell of type cell at which to place the new card on the gameboard.
		 *  \return The CardStackT representation of the given cell.
		 *  \throws invalid_cell One or both of the cells provided are not valid cells on the gameboard.
		 */			
		CardStackT getCell(CellT cell, unsigned int n);

		/**
		 *  \brief Return whether a valid move exists.
		 *  \details Returns whether the current game state contains a valid move. This means that it is possible
		 *  to move a card from a given cell to any other cell on the board. Note that it is possible to have a valid move
		 *  but still be in an unwinnable game (e.g. the card can be moved back and forth between two Tableau cell).
		 *  This function does not check for this condition.
		 *  \return true if a valid move exists on the gameboard, and false if no valid move exists.
		 */
		bool validMoveExists();

		/**
		 *  \brief Return whether the gameboard is currently in the win state.
		 *  \details Returns whether the current game state is a win state. This would mean that all HomeCells
		 *  are stacked from Ace to King of each suit.
		 *  \return true if the gameboard is currently in a winning state, false otherwise.
		 */
		bool isInWinState();
};

#endif