/**
 * \file Stack.h
 * \brief A template module for a stack of a given type of elements.
 * \author Christopher W. Schankula, MacID: schankuc
 * \date April 8th, 2018
*/


#ifndef STACK_H
#define STACK_H

#include <vector>

/**
 *  \brief Template class for storing values in a first in, last out (FILO) stack.
 *  \details Implements a template for a first in, last out (FILO) stack.
 */
template<class T>
class Stack{
	private:
		std::vector<T> S;
		unsigned int max_size;

	public:
		/**
		 *  \brief Construct a new stack.
		 *  \details Create a new instance of stack given some initial elements and a maximum size.
		 *  \param n The maximum size of the given stack.
		 *  \param s The initial elements to push on to the stack. Elements are pushed in an order where 
		 *  the first elements (i.e s[0]) are pushed on to the stack first and thus would be the last to be popped.
		 *  \throw invalid_argument The number of elements in the initial stack given is greater than
		 the maximum stack size.
		 */
		Stack(unsigned int n, std::vector<T> s);

		/**
		 *  \brief Creates a new stack by pushing the given value on to the current stack.
		 *  \details Create a new instance of stack identical to the last one except with one more value, the one given,
		 * placed on top.
		 *  \param e The new element of type T to place on top of the stack.
		 *  \return A new instance of stack with the same elements as the current one, with the new value e on the top
		 * of the stack.
		 *  \throw stack_full The number of elements on the current stack already equals the maximum
		 * capacity of the stack.
		 */
		Stack<T> push(T e);

		/**
		 *  \brief Creates a new stack by popping the top value off the current one.
		 *  \details Create a new instance of stack identical to the last one except with one less value:
		 * the topmost value on the stack is deleted.
		 *  \return A new instance of stack with one less value than the current one, the top having been 
		 * removed.
		 *  \throw stack_empty The current stack does not have any more values to pop off.
		 */
		Stack<T> pop();

		/**
		 *  \brief Get the top value of the current stack.
		 *  \details Returns the topmost value on the current stack, the value most recently added to the stack.
		 *  \return The top value of type T of the current stack.
		 *  \throw stack_empty The current stack does not have values to return.
		 */		
		T top();

		/**
		 *  \brief Get the current size of the stack.
		 *  \details Returns the number of elements currently stored on the current stack.
		 *  \return The number of elements on the current stack.
		 */
		unsigned int getSize();

		/**
		 *  \brief Get the maximum size of the stack.
		 *  \details Returns the maximum number of elements of type T which can be stored on the current stack.
		 *  \return The maximum number of elements which can be stored the current stack.
		 */
		unsigned int getMaxSize();

		/**
		 *  \brief Get a sequence representation of the current stack.
		 *  \details Returns a sequence representing the elements in the current stack. The beginning of the sequence
		 *  represents the bottommost values on the stack.
		 *  \return The sequence representation of the current stack.
		 */
		std::vector<T> toSeq();
};

#endif