/**
 * \file GameBoard.cpp
 * \brief The implementation of the Freecell gameboard ADT described in GameBoard.h.
 * \author Christopher W. Schankula, MacID: schankuc
 * \date April 8th, 2018
*/

#include "GameBoard.h"
#include "CardStack.h"
#include "CardTypes.h"
#include "GameBoardTypes.h"
#include "Exceptions.h"
#include <stddef.h>

GameBoardT::GameBoardT(){
	this->T = GameBoardT::stackSeq(8,19);
	this->F = GameBoardT::stackSeq(4,1);
	this->H = GameBoardT::stackSeq(4,13);
}

GameBoardT::GameBoardT(std::vector<CardT> deck){
	if(deck.size() != 52) throw invalid_argument();

	this->T = GameBoardT::tableauDeck(deck);
	this->F = GameBoardT::stackSeq(4,1);
	this->H = GameBoardT::stackSeq(4,13);
}

void GameBoardT::place(CellT cell, unsigned int i, CardT card){
	switch (cell){
		case Tableau:
			if ((i < 8) && (this->T[i].getSize() == this->T[i].getMaxSize()))
				throw invalid_config();
			else if (i >= 8)
				throw invalid_cell();
			break;
		case FreeCell:
			if ((i < 4) && (this->F[i].getSize() == this->F[i].getMaxSize()))
				throw invalid_config();
			else if (i >= 4)
				throw invalid_cell();
			break;
		case HomeCell:
			if ((i < 4) && (this->H[i].getSize() == this->H[i].getMaxSize()))
				throw invalid_config();
			else if (i >= 4)
				throw invalid_cell();
			break;
	}

	switch (cell){
		case Tableau:
			this->T[i] = this->T[i].push(card);
			break;
		case FreeCell:
			this->F[i] = this->F[i].push(card);
			break;
		case HomeCell:
			this->H[i] = this->H[i].push(card);
			break;
	}
}

bool GameBoardT::isValidMove(CellT c0, unsigned int n0, CellT c1, unsigned int n1){
	if (!isValidCell(c0, n0) || !isValidCell(c1, n1)) throw invalid_cell();

	switch (c0){
		case Tableau:
			switch (c1){
				case Tableau:
					return validTabTab(n0, n1);
				case FreeCell:
					return validTabFree(n0, n1);
				case HomeCell:
					return validTabHome(n0, n1);
			}
		case FreeCell:
			switch (c1){
				case Tableau:
					return validFreeTab(n0, n1);
				case FreeCell:
					return validFreeFree(n0, n1);
				case HomeCell:
					return validFreeHome(n0, n1);
			}
		case HomeCell:
			return false;
	}
}

void GameBoardT::makeMove(CellT c0, unsigned int n0, CellT c1, unsigned int n1){
	if (!isValidMove(c0, n0, c1, n1)) throw invalid_move();
	if (!isValidCell(c0, n0) || !isValidCell(c1, n1)) throw invalid_cell();

	switch (c0){
		case Tableau:
			switch (c1){
				case Tableau:
					this->T[n1] = this->T[n1].push(this->T[n0].top());
					this->T[n0] = this->T[n0].pop();
					return;
				case FreeCell:
					this->F[n1] = this->F[n1].push(this->T[n0].top());
					this->T[n0] = this->T[n0].pop();
					return;
				case HomeCell:
					this->H[n1] = this->H[n1].push(this->T[n0].top());
					this->T[n0] = this->T[n0].pop();
					return;
			}
		case FreeCell:
			switch (c1){
				case Tableau:
					this->T[n1] = this->T[n1].push(this->F[n0].top());
					this->F[n0] = this->F[n0].pop();
					return;
				case FreeCell:
					this->F[n1] = this->F[n1].push(this->F[n0].top());
					this->F[n0] = this->F[n0].pop();
					return;
				case HomeCell:
					this->H[n1] = this->H[n1].push(this->F[n0].top());
					this->F[n0] = this->F[n0].pop();
					return;
			}
		case HomeCell: return;
	}
}

CardStackT GameBoardT::getCell(CellT cell, unsigned int n){
	if (!isValidCell(cell, n)) throw invalid_cell();

	switch (cell){
		case Tableau:
			return this->T[n];
		case FreeCell:
			return this->F[n];
		case HomeCell:
			return this->H[n];
	}
}

bool GameBoardT::validMoveExists(){
	CellT cells[] = {Tableau, FreeCell, HomeCell};
	unsigned int size[] = {8, 4, 4};

	for(int i = 0; i < 3; i++){
		CellT c0 = cells[i];
		for (int n0 = 0; n0 < size[i]; n0++){
			for (int j = 0; j < 3; j++){
				CellT c1 = cells[j];
				for(int n1 = 0; n1 < size[j]; n1++){
					if (isValidMove(c0, n0, c1, n1)) return true;
				}
			}
		}
	}
	return false;
}

bool GameBoardT::isInWinState(){
	return isWinningHomeCell(0) && isWinningHomeCell(1) && isWinningHomeCell(2) && isWinningHomeCell(3);
}

/****** private local functions ******/

std::vector<CardStackT> GameBoardT::stackSeq(unsigned int l, unsigned int n){
	std::vector<CardStackT> stSeq = std::vector<CardStackT>();

	for(int i = 0; i < l; i++){
		stSeq.push_back(CardStackT(n, std::vector<CardT>()));
	}

	return stSeq;
}


std::vector<CardStackT> GameBoardT::tableauDeck(std::vector<CardT> deck){
	std::vector<CardStackT> stSeq = std::vector<CardStackT>();

	unsigned int lengths[] = {7, 7, 7, 7, 6, 6, 6, 6};
	unsigned int sizes[] = {19,19,19,19,18,18,18,18};

	int n = 0;
	for (int i = 0; i < 8; i++){
		std::vector<CardT> slice(deck.begin() + n, deck.begin() + n + lengths[i]);
		stSeq.push_back(CardStackT(sizes[i], slice));
		n += lengths[i];
	}

	return stSeq;
}

bool GameBoardT::isValidCell(CellT c, unsigned int n){
	switch (c){
		case Tableau:
			return n < 8;
		case FreeCell:
			return n < 4;
		case HomeCell:
			return n < 4;
	}
}

bool GameBoardT::validTabTab(unsigned int t0, unsigned int t1){
	if (this->T[t0].getSize() > 0){
		if(this->T[t1].getSize() > 0)	
			return tabPlaceable(this->T[t0].top(), this->T[t1].top());
		else if (this->T[t1].getSize() == 0) 
			return true;
	}
	return false;
}

bool GameBoardT::validTabFree(unsigned int t, unsigned int f){
	if (this->T[t].getSize() > 0){
		if(this->F[f].getSize() > 0)	
			return false;
		else if (this->F[f].getSize() == 0) 
			return true;
	}
	return false;
}

bool GameBoardT::validTabHome(unsigned int t, unsigned int h){
	if (this->T[t].getSize() > 0){
		if(H[h].getSize() > 0)	
			return homePlaceable(T[t].top(), H[h].top());
		else if (this->H[h].getSize() == 0) 
			return this->T[t].top().getRank() == A;
	}
	return false;
}

bool GameBoardT::validFreeTab(unsigned int f, unsigned int t){
	if (this->F[f].getSize() > 0){
		if(this->T[t].getSize() > 0)	
			return tabPlaceable(this->F[f].top(), this->T[t].top());
		else if (this->T[t].getSize() == 0) 
			return true;
	}
	return false;
}

bool GameBoardT::validFreeFree(unsigned int f0, unsigned int f1){
	return (this->F[f0].getSize() > 0 && this->F[f1].getSize() == 0); 
}

bool GameBoardT::validFreeHome(unsigned int f, unsigned int h){
	if (this->F[f].getSize() > 0){
		if(this->H[h].getSize() > 0)	
			return homePlaceable(F[f].top(), this->H[h].top());
		else if (this->H[h].getSize() == 0) 
			return this->F[f].top().getRank() == A;
	}
	return false;
}

bool GameBoardT::oneLess(CardT c0, CardT c1){
	RankT ranks[] = {A, C2, C3, C4, C5, C6, C7, C8, C9, C10, J, Q, K};

	for(int i = 0; i < 12; i++)
		if (c0.getRank() == ranks[i] && c1.getRank() == ranks[i + 1]) return true;
	return false;
}

bool GameBoardT::oneMore(CardT c0, CardT c1){
	RankT ranks[] = {A, C2, C3, C4, C5, C6, C7, C8, C9, C10, J, Q, K};

	for(int i = 0; i < 12; i++)
		if (c1.getRank() == ranks[i] && c0.getRank() == ranks[i + 1]) return true;
	return false;
}

bool GameBoardT::tabPlaceable(CardT c0, CardT c1){
	if (c0.getColour() != c1.getColour()) return oneLess(c0, c1);
	else return false;
}

bool GameBoardT::homePlaceable(CardT c0, CardT c1){
	if (c0.getSuit() == c1.getSuit()) return oneMore(c0, c1);
	else return false;
}

bool GameBoardT::isWinningHomeCell(unsigned int i){
	std::vector<CardT> stSeq = H[i].toSeq();

	if (!(stSeq.size() == 13)) return false;
	if (!(stSeq[0].getRank() == A)) return false;

	bool winning = true;

	for (int j = 0; j < H[i].getSize() - 1; j++)
		winning = winning && oneLess(stSeq[j], stSeq[j + 1]);

	return winning;
}
