/**
 * \file testCardADT.cpp
 * \brief Implementation of CardT tests.
 * \author Christopher W. Schankula, MacID: schankuc
 * \date April 9th, 2018
*/


/*
* NOTABLE TEST CASES (ACCORDING TO RUBRIC)
* Lines 135-163: Test case for an invalid move between columns
* Lines 165-175, 301-315: Test case for moving to free cell, when one is available
* Lines 177-215, 317-350: Test case for moving to free cell, when no free cell available
* Lines 804-818: Test case for detecting winning game board (true case and false case)
* Lines 756-801: Test case for detecting no valid moves left (true case and false case)
*/

#include "catch.h"
#include "GameBoard.h"
#include "GameBoardTypes.h"
#include "CardDeckUtils.h"
#include <iostream>
#include "Exceptions.h"

TEST_CASE( "tests for GameBoard" , "[GameBoard]" ) {
	GameBoardT gb = GameBoardT();

	SECTION("test GameBoardT blank constructor"){
		CellT cellTypes[] = {Tableau, FreeCell, HomeCell};
		unsigned int cellLens[] = {8, 4, 4};

		//require that all cells are empty at the start
		for (int i = 0; i < 3; i++){
			CellT c = cellTypes[i];
			for (int j = 0; j < cellLens[i]; j++)
				REQUIRE(gb.getCell(c, i).getSize() == 0);
		}
	}

	SECTION("test GameBoardT deck constructor"){
		CellT cellTypes[] = {Tableau, FreeCell, HomeCell};
		unsigned int cellLens[] = {8, 4, 4};
		//correct sizes of cell 						    //sizes of tableau spaces  //FreeCells and HomeCells empty (below)
		std::vector<std::vector<unsigned int>> cellSizes = {{7, 7, 7, 7, 6, 6, 6, 6}, {0, 0, 0, 0}, {0, 0, 0, 0}};

		std::vector<CardT> deck = CardDeckUtils::randomDeck();
		gb = GameBoardT(deck);

		//require that all cells are the correct sizes at the start
		for (int i = 0; i < 3; i++){
			CellT c = cellTypes[i];
			for (int j = 0; j < cellLens[i]; j++)
				REQUIRE(gb.getCell(c, j).getSize() == cellSizes[i][j]);
		}

		//require that the top cards of each Tableau space are in the correct places according to the spec
		int k = 0;
		for (int j = 0; j < cellLens[0]; j++){
			k += cellSizes[0][j];
			REQUIRE(gb.getCell(Tableau, j).top().getRank() == deck[k-1].getRank());
		}
	}

	SECTION("test GameBoardT deck constructor throws invalid_argument"){
		CardT c0 = CardT(H, C10);

		std::vector<CardT> singleCard = std::vector<CardT>();

		singleCard.push_back(c0);

		REQUIRE_THROWS_AS(GameBoardT(singleCard), invalid_argument);
	}


	SECTION("test place() places card correctly"){
		CardT c0 = CardT(H, C10);

		gb.place(Tableau, 0, c0);

		REQUIRE(gb.getCell(Tableau, 0).top().getSuit() == H);
		REQUIRE(gb.getCell(Tableau, 0).top().getRank() == C10);
	}
	
	SECTION("test place() throws invalid_config if trying to place card incorrectly"){
		CardT c0 = CardT(H, C10);

		for (int i = 0; i < 19; i++){
			REQUIRE(gb.getCell(Tableau, 0).getSize() == i);
			gb.place(Tableau, 0, c0);
		}

		gb.place(FreeCell, 0, c0);
		REQUIRE(gb.getCell(FreeCell, 0).getSize() == 1);

		for (int i = 0; i < 13; i++){
			REQUIRE(gb.getCell(HomeCell, 0).getSize() == i);
			gb.place(HomeCell, 0, c0);
		}

		REQUIRE_THROWS_AS(gb.place(Tableau, 0, c0), invalid_config);
		REQUIRE_THROWS_AS(gb.place(FreeCell, 0, c0), invalid_config);
		REQUIRE_THROWS_AS(gb.place(HomeCell, 0, c0), invalid_config);
	}

	SECTION("test place() throws invalid_cell if trying to place card on incorrect cell"){
		CardT c0 = CardT(H, C10);

		REQUIRE_THROWS_AS(gb.place(Tableau, 8, c0), invalid_cell);
		REQUIRE_THROWS_AS(gb.place(FreeCell, 5, c0), invalid_cell);
		REQUIRE_THROWS_AS(gb.place(HomeCell, 5, c0), invalid_cell);
	}

	SECTION("test getCell() gets correct cell"){
		CardT c0 = CardT(H, C10);

		gb.place(Tableau, 0, c0);

		REQUIRE(gb.getCell(Tableau, 0).top().getRank() == C10);
		REQUIRE(gb.getCell(Tableau, 0).top().getSuit() == H);
	}

	SECTION("test getCell() throws invalid_cell"){
		REQUIRE_THROWS_AS(gb.getCell(Tableau, 9), invalid_cell);
		REQUIRE_THROWS_AS(gb.getCell(HomeCell, 4), invalid_cell);
		REQUIRE_THROWS_AS(gb.getCell(FreeCell, 4), invalid_cell);
	}

                           /******* TESTS FOR isValidMove() *******/

	SECTION("test isValidMove() throws invalid_cell"){
		REQUIRE_THROWS_AS(gb.isValidMove(Tableau, 9, FreeCell, 0), invalid_cell);
		REQUIRE_THROWS_AS(gb.isValidMove(Tableau, 0, FreeCell, 5), invalid_cell);
	}

	SECTION("test isValidMove() for moving a card from a Tableau to a Tableau - valid moves"){
		CardT c0 = CardT(H, J);
		CardT c1 = CardT(S, C10);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);

		//This is a valid move (black 10 on to red Jack)
		REQUIRE(gb.isValidMove(Tableau, 1, Tableau, 0));

		//This is a valid move (black 10 on to empty Tableau cell 2)
		REQUIRE(gb.isValidMove(Tableau, 1, Tableau, 2));
	}	

	SECTION("test isValidMove() for moving a card from a Tableau to a Tableau - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);

		//This is a not a valid move (red Jack on to black 10)
		REQUIRE(!gb.isValidMove(Tableau, 0, Tableau, 1));

		//This is a not a valid move (Tableau 2 is empty cell)
		REQUIRE(!gb.isValidMove(Tableau, 2, Tableau, 1));
	}

	SECTION("test isValidMove() for moving a card from a Tableau to a FreeCell - valid move"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);

		//This is a valid move (red Jack on to empty freecell)
		REQUIRE(gb.isValidMove(Tableau, 0, FreeCell, 0));
	}	

	SECTION("test isValidMove() for moving a card from a Tableau to a FreeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);
		CardT c2 = CardT(C, A);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);
		//place card on FreeCell
		gb.place(FreeCell, 0, c2);

		//This is a not a valid move (empty Tableau cell to empty freecell)
		REQUIRE(!gb.isValidMove(Tableau, 2, FreeCell, 0));

		//This is a not a valid move (red Jack to occupied freecell)
		REQUIRE(!gb.isValidMove(Tableau, 0, FreeCell, 0));
	}

	SECTION("test isValidMove() for moving a card from a Tableau to a FreeCell - no freecells available"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(H, C5);
		CardT c4 = CardT(D, K);

		//place cards on all four freecells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(FreeCell, 2, c2);
		gb.place(FreeCell, 3, c3);
		//place a card on a tableau cell to move
		gb.place(Tableau, 0, c4);

		//These are not valid moves (all freecells used)
		REQUIRE(!gb.isValidMove(Tableau, 0, FreeCell, 0));
		REQUIRE(!gb.isValidMove(Tableau, 0, FreeCell, 1));
		REQUIRE(!gb.isValidMove(Tableau, 0, FreeCell, 2));
		REQUIRE(!gb.isValidMove(Tableau, 0, FreeCell, 3));
	}	

	SECTION("test isValidMove() for moving a card from a Tableau to a HomeCell - valid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, A);
		CardT c2 = CardT(S, C2);
		CardT c3 = CardT(S, C3);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c3);
		gb.place(Tableau, 2, c2);
		gb.place(Tableau, 1, c1);

		//This is a valid move (ace to empty HomeCell)
		REQUIRE(gb.isValidMove(Tableau, 1, HomeCell, 0));
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 1, HomeCell, 0));

		//This is a valid move (2 to ace of same suit)
		REQUIRE(gb.isValidMove(Tableau, 2, HomeCell, 0));
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 2, HomeCell, 0));

		//This is a valid move (2 to 3 of same suit)
		REQUIRE(gb.isValidMove(Tableau, 1, HomeCell, 0));
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 1, HomeCell, 0));
	}	

	SECTION("test isValidMove() for moving a card from a Tableau to a HomeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C2);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(C, C3);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);
		//place card on FreeCell
		gb.place(Tableau, 3, c3);
		gb.place(HomeCell, 0, c2);

		//This is a not a valid move (empty Tableau cell to empty homecell)
		REQUIRE(!gb.isValidMove(Tableau, 2, HomeCell, 1));

		//This is a not a valid move (3 of Clubs to ace of clubs)
		REQUIRE(!gb.isValidMove(Tableau, 3, HomeCell, 0));

		//This is a not a valid move (2 of spades to ace of clubs)
		REQUIRE(!gb.isValidMove(Tableau, 1, HomeCell, 0));
	}

	//moving from a FreeCell to any type of cell


	SECTION("test isValidMove() for moving a card from a FreeCell to a Tableau - valid moves"){
		CardT c0 = CardT(H, J);
		CardT c1 = CardT(S, C10);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);

		//This is a valid move (jack of Hearts (red) on to empty tableau cell)
		REQUIRE(gb.isValidMove(FreeCell, 0, Tableau, 0));
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 0, Tableau, 0));

		//This is a valid move (black 10 on to red jack of hearts)
		REQUIRE(gb.isValidMove(FreeCell, 1, Tableau, 0));
	}	

	SECTION("test isValidMove() for moving a card from a FreeCell to a Tableau - invalid moves"){
		CardT c0 = CardT(H, J);
		CardT c1 = CardT(S, C10);
		CardT c2 = CardT(S, J);

		//place cards on freecell and tableau cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(Tableau, 0, c2);

		//This is an invalid move (empty FreeCell to tableau cell)
		REQUIRE(!gb.isValidMove(FreeCell, 2, Tableau, 0));

		//This is an invalid move (10 of Spades on to Jack of Spades)
		REQUIRE(!gb.isValidMove(FreeCell, 1, Tableau, 0));
	}

	SECTION("test isValidMove() for moving a card from a FreeCell to a FreeCell - valid move"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);

		//This is a valid move (red 10 on to empty freecell)
		REQUIRE(gb.isValidMove(FreeCell, 0, FreeCell, 2));
		REQUIRE(gb.isValidMove(FreeCell, 0, FreeCell, 3));
		//This is a valid move (black 9 to empty freecell)
		REQUIRE(gb.isValidMove(FreeCell, 1, FreeCell, 3));
		REQUIRE(gb.isValidMove(FreeCell, 0, FreeCell, 3));
	}	

	SECTION("test isValidMove() for moving a card from a FreeCell to a FreeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);

		//This is an invalid move (empty freecell to other empty freecell)
		REQUIRE(!gb.isValidMove(FreeCell, 2, FreeCell, 3));
		REQUIRE(!gb.isValidMove(FreeCell, 3, FreeCell, 2));
		//This is an invalid move (freecell with a card to an occupied one)
		REQUIRE(!gb.isValidMove(FreeCell, 1, FreeCell, 0));
		REQUIRE(!gb.isValidMove(FreeCell, 0, FreeCell, 1));
	}	

	SECTION("test isValidMove() for moving a card from a FreeCell to a FreeCell - no freecells available"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(H, C5);

		//place cards on all four freecells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(FreeCell, 2, c2);
		gb.place(FreeCell, 3, c3);
		//place a card on a tableau cell to move

		//These are not valid moves (all freecells used)
		REQUIRE(!gb.isValidMove(FreeCell, 0, FreeCell, 1));
		REQUIRE(!gb.isValidMove(FreeCell, 0, FreeCell, 2));
		REQUIRE(!gb.isValidMove(FreeCell, 0, FreeCell, 3));
	}	

	SECTION("test isValidMove() for moving a card from a FreeCell to a HomeCell - valid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, A);
		CardT c2 = CardT(S, C2);
		CardT c3 = CardT(S, C3);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(FreeCell, 2, c2);
		gb.place(FreeCell, 3, c3);

		//This is a valid move (ace to empty HomeCell)
		REQUIRE(gb.isValidMove(FreeCell, 1, HomeCell, 0));
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 1, HomeCell, 0));

		//This is a valid move (2 to ace of same suit)
		REQUIRE(gb.isValidMove(FreeCell, 2, HomeCell, 0));
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 2, HomeCell, 0));

		//This is a valid move (2 to 3 of same suit)
		REQUIRE(gb.isValidMove(FreeCell, 3, HomeCell, 0));
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 3, HomeCell, 0));
	}	

	SECTION("test isValidMove() for moving a card from a FreeCell to a HomeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C2);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(C, C3);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		//place card on FreeCell
		gb.place(FreeCell, 3, c3);
		gb.place(HomeCell, 0, c2);

		//This is a not a valid move (empty FreeCell to empty homecell)
		REQUIRE(!gb.isValidMove(FreeCell, 2, HomeCell, 1));

		//This is a not a valid move (3 of Clubs to ace of clubs)
		REQUIRE(!gb.isValidMove(FreeCell, 3, HomeCell, 0));

		//This is a not a valid move (2 of spades to ace of clubs)
		REQUIRE(!gb.isValidMove(FreeCell, 1, HomeCell, 0));
	}

	SECTION("moves from from HomeCell are invalid"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C2);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(C, C3);

		gb.place(HomeCell, 0, c2);

		CellT cellTypes[] = {Tableau, HomeCell, FreeCell};

		//try moving from all HomeCells to all other types of cells
		for (int i = 0; i < 4; i++)
			for(int j = 0; j < 3; j++)
				REQUIRE(!gb.isValidMove(HomeCell, i, cellTypes[j], 0));
	}


                           /******* TESTS FOR makeMove() *******/

	SECTION("test makeMove() for moving a card from a Tableau to a Tableau - valid moves"){
		CardT c0 = CardT(H, J);
		CardT c1 = CardT(S, C10);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);

		REQUIRE(gb.getCell(Tableau, 0).getSize() == 1);
		REQUIRE(gb.getCell(Tableau, 1).getSize() == 1);

		//This is a valid move (black 10 on to red Jack)
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 1, Tableau, 0));

		//verify correct state afterwards
		REQUIRE(gb.getCell(Tableau, 1).getSize() == 0);
		REQUIRE(gb.getCell(Tableau, 0).getSize() == 2);
		REQUIRE(gb.getCell(Tableau, 0).top().getRank() == C10);
		REQUIRE(gb.getCell(Tableau, 0).top().getSuit() == S);
	}	

	SECTION("test makeMove() for moving a card from a Tableau to a Tableau - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);

		//This is a not a valid move (red Jack on to black 10)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 0, Tableau, 1), invalid_move);

		//This is a not a valid move (Tableau 2 is empty cell)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 2, Tableau, 1), invalid_move);
	}

	SECTION("test makeMove() for moving a card from a Tableau to a FreeCell - valid move"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);


		REQUIRE(gb.getCell(Tableau, 0).getSize() == 1);
		REQUIRE(gb.getCell(Tableau, 1).getSize() == 1);

		//This is a valid move (red Jack on to empty freecell)
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 0, FreeCell, 0));

		REQUIRE(gb.getCell(Tableau, 0).getSize() == 0);
		REQUIRE(gb.getCell(FreeCell, 0).getSize() == 1);
		REQUIRE(gb.getCell(FreeCell, 0).top().getRank() == C10);
		REQUIRE(gb.getCell(FreeCell, 0).top().getSuit() == H);
	}	

	SECTION("test makeMove() for moving a card from a Tableau to a FreeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);
		CardT c2 = CardT(C, A);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);
		//place card on FreeCell
		gb.place(FreeCell, 0, c2);

		//This is a not a valid move (empty Tableau cell to empty freecell)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 2, FreeCell, 0), invalid_move);

		//This is a not a valid move (red Jack to occupied freecell)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 0, FreeCell, 0), invalid_move);
	}

	SECTION("test makeMove() for moving a card from a Tableau to a FreeCell - no freecells available"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(H, C5);
		CardT c4 = CardT(D, K);

		//place cards on all four freecells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(FreeCell, 2, c2);
		gb.place(FreeCell, 3, c3);
		//place a card on a tableau cell to move
		gb.place(Tableau, 0, c4);

		//These are not valid moves (all freecells used)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 0, FreeCell, 0), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 0, FreeCell, 1), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 0, FreeCell, 2), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 0, FreeCell, 3), invalid_move);
	}	

	SECTION("test makeMove() for moving a card from a Tableau to a HomeCell - valid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, A);
		CardT c2 = CardT(S, C2);
		CardT c3 = CardT(S, C3);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c3);
		gb.place(Tableau, 2, c2);
		gb.place(Tableau, 1, c1);

		//This is a valid move (ace to empty HomeCell)
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 1, HomeCell, 0));

		REQUIRE(gb.getCell(HomeCell, 0).getSize() == 1);
		REQUIRE(gb.getCell(Tableau, 1).getSize() == 1);
		REQUIRE(gb.getCell(HomeCell, 0).top().getRank() == A);
		REQUIRE(gb.getCell(HomeCell, 0).top().getSuit() == S);

		//This is a valid move (2 to ace of same suit)
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 2, HomeCell, 0));

		REQUIRE(gb.getCell(HomeCell, 0).getSize() == 2);
		REQUIRE(gb.getCell(Tableau, 2).getSize() == 0);
		REQUIRE(gb.getCell(HomeCell, 0).top().getRank() == C2);
		REQUIRE(gb.getCell(HomeCell, 0).top().getSuit() == S);

		//This is a valid move (2 to 3 of same suit)
		REQUIRE_NOTHROW(gb.makeMove(Tableau, 1, HomeCell, 0));

		REQUIRE(gb.getCell(HomeCell, 0).getSize() == 3);
		REQUIRE(gb.getCell(Tableau, 1).getSize() == 0);
		REQUIRE(gb.getCell(HomeCell, 0).top().getRank() == C3);
		REQUIRE(gb.getCell(HomeCell, 0).top().getSuit() == S);
	}	

	SECTION("test makeMove() for moving a card from a Tableau to a HomeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C2);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(C, C3);

		//place cards on tableaux cells
		gb.place(Tableau, 0, c0);
		gb.place(Tableau, 1, c1);
		//place card on FreeCell
		gb.place(Tableau, 3, c3);
		gb.place(HomeCell, 0, c2);

		//This is a not a valid move (empty Tableau cell to empty homecell)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 2, HomeCell, 1), invalid_move);

		//This is a not a valid move (3 of Clubs to ace of clubs)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 3, HomeCell, 0), invalid_move);

		//This is a not a valid move (2 of spades to ace of clubs)
		REQUIRE_THROWS_AS(gb.makeMove(Tableau, 1, HomeCell, 0), invalid_move);
	}

	//moving from a FreeCell to any type of cell


	SECTION("test makeMove() for moving a card from a FreeCell to a Tableau - valid moves"){
		CardT c0 = CardT(H, J);
		CardT c1 = CardT(S, C10);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);

		//This is a valid move (jack of Hearts (red) on to empty tableau cell)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 0, Tableau, 0));

		REQUIRE(gb.getCell(FreeCell, 0).getSize() == 0);
		REQUIRE(gb.getCell(Tableau, 0).getSize() == 1);
		REQUIRE(gb.getCell(Tableau, 0).top().getRank() == J);
		REQUIRE(gb.getCell(Tableau, 0).top().getSuit() == H);

		//This is a valid move (black 10 on to red jack of hearts)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 1, Tableau, 0));

		REQUIRE(gb.getCell(FreeCell, 1).getSize() == 0);
		REQUIRE(gb.getCell(Tableau, 0).getSize() == 2);
		REQUIRE(gb.getCell(Tableau, 0).top().getRank() == C10);
		REQUIRE(gb.getCell(Tableau, 0).top().getSuit() == S);
	}	

	SECTION("test makeMove() for moving a card from a FreeCell to a Tableau - invalid moves"){
		CardT c0 = CardT(H, J);
		CardT c1 = CardT(S, C10);
		CardT c2 = CardT(S, J);

		//place cards on freecell and tableau cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(Tableau, 0, c2);

		//This is an invalid move (empty FreeCell to tableau cell)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 2, Tableau, 0), invalid_move);

		//This is an invalid move (10 of Spades on to Jack of Spades)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 1, Tableau, 0), invalid_move);
	}

	SECTION("test makeMove() for moving a card from a FreeCell to a FreeCell - valid move"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);

		//This is a valid move (red 10 on to empty freecell)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 0, FreeCell, 2));

		REQUIRE(gb.getCell(FreeCell, 0).getSize() == 0);
		REQUIRE(gb.getCell(FreeCell, 2).getSize() == 1);
		REQUIRE(gb.getCell(FreeCell, 2).top().getRank() == C10);
		REQUIRE(gb.getCell(FreeCell, 2).top().getSuit() == H);
		//This is a valid move (black 9 to empty freecell)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 1, FreeCell, 3));

		REQUIRE(gb.getCell(FreeCell, 1).getSize() == 0);
		REQUIRE(gb.getCell(FreeCell, 3).getSize() == 1);
		REQUIRE(gb.getCell(FreeCell, 3).top().getRank() == C9);
		REQUIRE(gb.getCell(FreeCell, 3).top().getSuit() == S);
	}	

	SECTION("test makeMove() for moving a card from a FreeCell to a FreeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);

		//This is an invalid move (empty freecell to other empty freecell)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 2, FreeCell, 3), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 3, FreeCell, 2), invalid_move);
		//This is an invalid move (freecell with a card to an occupied one)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 1, FreeCell, 0), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 0, FreeCell, 1), invalid_move);
	}	

	SECTION("test makeMove() for moving a card from a FreeCell to a FreeCell - no freecells available"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C9);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(H, C5);

		//place cards on all four freecells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(FreeCell, 2, c2);
		gb.place(FreeCell, 3, c3);
		//place a card on a tableau cell to move

		//These are not valid moves (all freecells used)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 0, FreeCell, 1), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 0, FreeCell, 2), invalid_move);
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 0, FreeCell, 3), invalid_move);
	}	

	SECTION("test makeMove() for moving a card from a FreeCell to a HomeCell - valid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, A);
		CardT c2 = CardT(S, C2);
		CardT c3 = CardT(S, C3);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		gb.place(FreeCell, 2, c2);
		gb.place(FreeCell, 3, c3);

		//This is a valid move (ace to empty HomeCell)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 1, HomeCell, 0));

		REQUIRE(gb.getCell(FreeCell, 1).getSize() == 0);
		REQUIRE(gb.getCell(HomeCell, 0).getSize() == 1);
		REQUIRE(gb.getCell(HomeCell, 0).top().getRank() == A);
		REQUIRE(gb.getCell(HomeCell, 0).top().getSuit() == S);

		//This is a valid move (2 to ace of same suit)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 2, HomeCell, 0));

		REQUIRE(gb.getCell(FreeCell, 2).getSize() == 0);
		REQUIRE(gb.getCell(HomeCell, 0).getSize() == 2);
		REQUIRE(gb.getCell(HomeCell, 0).top().getRank() == C2);
		REQUIRE(gb.getCell(HomeCell, 0).top().getSuit() == S);

		//This is a valid move (2 to 3 of same suit)
		REQUIRE_NOTHROW(gb.makeMove(FreeCell, 3, HomeCell, 0));

		REQUIRE(gb.getCell(FreeCell, 3).getSize() == 0);
		REQUIRE(gb.getCell(HomeCell, 0).getSize() == 3);
		REQUIRE(gb.getCell(HomeCell, 0).top().getRank() == C3);
		REQUIRE(gb.getCell(HomeCell, 0).top().getSuit() == S);
	}	

	SECTION("test makeMove() for moving a card from a FreeCell to a HomeCell - invalid moves"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C2);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(C, C3);

		//place cards on tableaux cells
		gb.place(FreeCell, 0, c0);
		gb.place(FreeCell, 1, c1);
		//place card on FreeCell
		gb.place(FreeCell, 3, c3);
		gb.place(HomeCell, 0, c2);

		//This is a not a valid move (empty FreeCell to empty homecell)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 2, HomeCell, 1), invalid_move);

		//This is a not a valid move (3 of Clubs to ace of clubs)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 3, HomeCell, 0), invalid_move);

		//This is a not a valid move (2 of spades to ace of clubs)
		REQUIRE_THROWS_AS(gb.makeMove(FreeCell, 1, HomeCell, 0), invalid_move);
	}

	SECTION("moves from from HomeCell are invalid"){
		CardT c0 = CardT(H, C10);
		CardT c1 = CardT(S, C2);
		CardT c2 = CardT(C, A);
		CardT c3 = CardT(C, C3);

		gb.place(HomeCell, 0, c2);

		CellT cellTypes[] = {Tableau, HomeCell, FreeCell};

		//try moving from all HomeCells to all other types of cells
		for (int i = 0; i < 4; i++)
			for(int j = 0; j < 3; j++)
				REQUIRE_THROWS_AS(gb.makeMove(HomeCell, i, cellTypes[j], 0), invalid_move);
	}

	SECTION("test no moves state - false (one FreeCell available)"){
		gb.place(Tableau, 0, CardT(C,J));
		gb.place(Tableau, 1, CardT(D, C3));
		gb.place(Tableau, 2, CardT(S, C7));
		gb.place(Tableau, 3, CardT(D, C2));
		gb.place(Tableau, 4, CardT(H, C4));
		gb.place(Tableau, 5, CardT(H, C7));
		gb.place(Tableau, 6, CardT(H, C2));
		gb.place(Tableau, 7, CardT(S, Q ));

		gb.place(FreeCell, 0, CardT(H, K ));
		gb.place(FreeCell, 1, CardT(D, C9));
		gb.place(FreeCell, 2, CardT(H, Q ));
		//i = 3 freecell available!
		gb.place(HomeCell, 0, CardT(C, C2));
		gb.place(HomeCell, 1, CardT(S, C3));

		REQUIRE(gb.validMoveExists());
	}

	SECTION("test no moves state - true (no cards at all)"){
		REQUIRE(!gb.validMoveExists());
	}

	SECTION("test no moves state - true (stuck game)"){
		//unwinnable game (only top cards in each stack matter)

		gb.place(Tableau, 0, CardT(C,J));
		gb.place(Tableau, 1, CardT(D, C3));
		gb.place(Tableau, 2, CardT(S, C7));
		gb.place(Tableau, 3, CardT(D, C2));
		gb.place(Tableau, 4, CardT(H, C4));
		gb.place(Tableau, 5, CardT(H, C7));
		gb.place(Tableau, 6, CardT(H, C2));
		gb.place(Tableau, 7, CardT(S, Q ));

		gb.place(FreeCell, 0, CardT(H, K ));
		gb.place(FreeCell, 1, CardT(D, C9));
		gb.place(FreeCell, 2, CardT(H, Q ));
		gb.place(FreeCell, 3, CardT(H, C8));
		gb.place(HomeCell, 0, CardT(C, C2));
		gb.place(HomeCell, 1, CardT(S, C3));
		//HomeCells i = 2 and i = 3 are empty

		REQUIRE(!gb.validMoveExists());
	}


	SECTION("test winning state - false"){
		REQUIRE(!gb.isInWinState());
	}

	SECTION("test winning state - true"){
		SuitT suits[] = {H, C, S, D};
		RankT ranks[] = {A, C2, C3, C4, C5, C6, C7, C8, C9, C10, J, Q, K};

		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 13; j++)
				gb.place(HomeCell, i, CardT(suits[i], ranks[j]));

		REQUIRE(gb.isInWinState());
	}
}
